#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import os
import logging
import unittest

# from pypushflow import UtilsMongoDb

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("testPythonActor")


class TestUtilsMongoDb(unittest.TestCase):
    def setUp(self):
        os.environ[
            "PYPUSHFLOW_MONGOURL"
        ] = "mongodb://pybes:pybes@linsvensson.esrf.fr:27017/pybes"
        os.environ["PYPUSHFLOW_CREATOR"] = "TestUtilsMongoDb"

    def tes_initMongo(self):
        name = "test_initMongo"
        # workflowId = UtilsMongoDb.initMongo(name=name)
        # self.assertIsNotNone(workflowId)

        # def tes_initActor(self):
        # name = 'test_initMongo'
        # workflowId = UtilsMongoDb.initMongo(name=name)
        # self.assertIsNotNone(workflowId)
        actorName = "TestActor"
        # actorId = UtilsMongoDb.initActor(name=actorName, workflowId=workflowId)
        # self.assertIsNotNone(actorId)

        # def test_addDataToActor(self):
        name = "test_initMongo"
        # workflowId = UtilsMongoDb.initMongo(name=name)
        # self.assertIsNotNone(workflowId)
        # actorName1 = 'TestActor1'
        # actorId1 = UtilsMongoDb.initActor(name=actorName1, workflowId=workflowId)
        # self.assertIsNotNone(actorId1)
        # actorName2 = 'TestActor2'
        # actorId2 = UtilsMongoDb.initActor(name=actorName2, workflowId=workflowId)
        # inData = {'a': 1}
        # UtilsMongoDb.addDataToActor(workflowId=workflowId, actorId=actorId1, actorData={'inData': inData})


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestUtilsMongoDb,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2019-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/04/2020"


class BaseWorkflowAddOn:
    """Base class for a Workflow instance"""

    def __init__(self, workflow_instance):
        pass


class BaseActorAddOn:
    """Base class to define an add on for an actor"""

    def __init__(self, target):
        """Apply an add on to classes that are instances of target class"""
        self.__target = target

    @property
    def target(self):
        return self.__target

    def pre_trigger_action(self, actor, in_data):
        pass

    def post_trigger_action(self, actor):
        pass



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "29/05/2017"

import functools
import logging
import traceback
import inspect
import importlib
from typing import Union
from importlib.machinery import SourceFileLoader

_logger = logging.getLogger(__name__)

global next_node_free_idF
next_node_free_id = 0


def get_next_node_free_id() -> int:
    global next_node_free_id
    _id = next_node_free_id
    next_node_free_id += 1
    return _id


def trace_unhandled_exceptions(func):
    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        try:
            out_data = func(*args, **kwargs)
        except Exception as e:
            _logger.exception(e)
            errorMessage = "{0}".format(e)
            traceBack = traceback.format_exc()
            return WorkflowException(
                msg=errorMessage, traceBack=traceBack, data=args[1]
            )
        return out_data

    return wrapped_func


class Node(object):
    """
    Node in the `.Scheme`. Will be associated to a core process.

    :param processing_pt: pointer to a class or a function or str defining the
                          callback. If the callback is a class then the handler
                          should be defined or the class should have a default
                          'process' function that will be called by default.
    :param int id: unique id of the node.
    :param dict properties: properties of the node
    :param str luigi_task: luigi task associate to this node
    """

    need_stop_join = False
    """flag to stop the node only when receive the 'stop' signal"""

    _JSON_PROCESS_PT = "process_pt"
    _JSON_ID = "id"
    _JSON_PROPERTIES = "properties"
    _JSON_ERROR_HANDLER = "error_handler"

    def __init__(
        self,
        processing_pt,
        id: Union[None, int] = None,
        properties: Union[None, dict] = None,
        error_handler=None,
    ):
        self.id = get_next_node_free_id() if id is None else id
        """int of the node id"""
        self.properties = properties or {}
        """dict of the node properties"""
        self.upstream_nodes = set()
        """Set of upstream nodes"""
        self.downstream_nodes = set()
        """Set of downstream nodes"""
        self.__process_instance = None
        """"""
        self._process_pt = processing_pt
        """process instance"""
        self._handlers = {}
        """handlers with link name as key and callback as value.
        The default handler is store under the 'None' value"""
        self._input_type_to_name = {}
        """link input type to a signal name"""
        self._output_type_to_name = {}
        """link output type to a signal name"""
        self._error_handler = error_handler
        self.out_data = None

    def get_input_channel_name(self, data_object):
        for dtype, channel_name in self._input_type_to_name.items():
            if isinstance(data_object, dtype):
                return channel_name
        return None

    def get_output_channel_name(self, data_object):
        for dtype, channel_name in self._output_type_to_name.items():
            if isinstance(data_object, dtype):
                return channel_name
        return None

    @property
    def handlers(self) -> dict:
        return self._handlers

    @property
    def process_pt(self):
        return self._process_pt

    @property
    def class_instance(self):
        return self.__process_instance

    def isfinal(self) -> bool:
        """

        :return: True if the node is at the end of a branch.
        :rtype: bool
        """
        return len(self.downstream_nodes) is 0

    def isstart(self) -> bool:
        """

        :return: True if the node does not requires any input
        :rtype: bool
        """
        return len(self.upstream_nodes) is 0

    def get_class_name_and_module_name(self) -> tuple:
        """
        :retruns: (class name, module name)
        """
        sname = self._process_pt.rsplit(".")
        if not (len(sname) > 1):
            raise ValueError(self._process_pt + " is not recognized as a valid name")
        class_name = sname[-1]
        del sname[-1]
        mod_name = ".".join(sname)
        return class_name, mod_name

    def load_handlers(self) -> None:
        """
        load handlers from the `processing_pt` defined.
        For callable it will always be `processing_pt`.
        But for not-callable class it will be class function defined in the
        `inputs` variable.

        :raises: ValueError if unable to find sme handlers in the classes
                 definition
        """
        self._handlers.clear()
        self._input_type_to_name.clear()
        self._output_type_to_name.clear()

        assert self._process_pt is not None
        if callable(self._process_pt):
            self.__process_instance = self._process_pt
            self._handlers[None] = self._process_pt
        else:
            if not type(self._process_pt) is str:
                raise ValueError(
                    "process_pt should be a callable or path to a class or function"
                )
            else:
                class_name, module_name = self.get_class_name_and_module_name()
                if module_name.endswith(".py"):
                    # warning: in this case the file should not have any relative
                    module = SourceFileLoader(module_name, module_name).load_module()
                else:
                    module = importlib.import_module(module_name)

                class_or_fct = getattr(module, class_name)
                if inspect.isclass(class_or_fct):
                    _logger.debug("instanciate " + str(class_or_fct))
                    self.__process_instance = class_or_fct()
                else:
                    self.__process_instance = class_or_fct
                if callable(self.__process_instance):
                    self._handlers[None] = self.__process_instance
                # manage the case where a class has several input handler
                if hasattr(self.__process_instance, "inputs"):
                    for input_ in self.__process_instance.inputs:
                        input_name, input_type, input_handler = input_[:3]
                        _logger.debug(
                            "[node: %s] add input_name: %s, "
                            "input_type: %s, input_handler: %s"
                            % (self._process_pt, input_name, input_type, input_handler)
                        )
                        if str(input_type) in self._input_type_to_name:
                            raise ValueError(
                                "Several input name found for the "
                                "same input type. This case is not managed."
                            )
                        self._input_type_to_name[input_type] = input_name
                        self._handlers[input_name] = input_handler
                        # self._handlers[input_name] = getattr(self.__process_instance, input_handler)
                if hasattr(self.__process_instance, "outputs"):
                    for output_ in self.__process_instance.outputs:
                        output_name, output_type, output_handler = output_[:3]
                        _logger.debug(
                            "[node: %s] add output_name: %s, "
                            "output_type: %s, output_handler: %s"
                            % (self._process_pt, input_name, input_type, input_handler)
                        )
                        if output_type in self._output_type_to_name:
                            raise ValueError(
                                "Several output name found for the "
                                "same output type. This case is not managed."
                            )
                        self._output_type_to_name[output_type] = output_name

        if len(self._handlers) == 0:
            raise ValueError(
                "Fail to init handlers, none defined for " + str(self._process_pt)
            )

    @staticmethod
    def execute(
        process_pt, properties: dict, input_name: str, input_data: object
    ) -> tuple:
        """
        Create an instance of a node with 'process_pt' and execute it with the
        given input_name, properties and input_data.

        :param str process_pt: name of the process to execute
         (can be a module.class name, or a module.function)
        :param dict properties: process properties
        :param str input_name: name of the input data
        :param input_data: input data :warning: Should be serializable

        :return: (output data type, output data)
                 :warning: Should be serializable
        """
        node = Node(processing_pt=process_pt, properties=properties)
        node.load_handlers()
        logging.info(
            "start execution of {0} with {1} through channel {2}"
            "".format(str(process_pt), input_data, input_name)
        )
        if hasattr(node.__process_instance, "set_properties"):
            node.__process_instance.set_properties(properties)

        if input_name in node.handlers:
            if type(node.handlers[input_name]) is str:
                out = getattr(node.__process_instance, node.handlers[input_name])(
                    input_data
                )
            else:
                out = node.handlers[input_name](input_data)
        elif None in node.handlers:
            if type(node.handlers[None]) is str:
                out = getattr(node.__process_instance, node.handlers[None])(input_data)
            else:
                out = node.__process_instance(input_data)
        else:
            err = '"{0}" channel is not managed by {1}'.format(
                input_name, node._process_pt
            )
            raise KeyError(err)

        # retrieve output channel
        if out is None:
            output_channel = None
        else:
            output_channel = node.get_output_channel_name(out)

        if hasattr(out, "to_dict"):
            return output_channel, out.to_dict()
        else:
            return output_channel, out

    def to_json(self) -> dict:
        """

        :return: json description of the node
        :rtype: dict
        """
        res = {
            self._JSON_PROCESS_PT: self.process_pt,
            self._JSON_ID: self.id,
            self._JSON_PROPERTIES: self.properties,
        }
        res.update(self._get_error_handler_json())
        return res

    def _get_error_handler_json(self):
        error_handler_json = (
            self._error_handler.to_json() if self._error_handler else {}
        )
        return {self._JSON_ERROR_HANDLER: error_handler_json}

    @staticmethod
    def load_node_info_from_json(json_data: dict) -> tuple:
        """
        load fom json stream the Node Information

        :param json_data: node description
        :return: node id, properties, pointer to the process to run
        :rtype: tuple
        """
        # load properties
        if Node._JSON_PROPERTIES not in json_data:
            _logger.error("Missing node properties in json description")
            _properties = None
        else:
            _properties = json_data[Node._JSON_PROPERTIES]
            assert type(_properties) is dict
        # load id
        if Node._JSON_ID not in json_data:
            _logger.error("Missing node id in json description")
            _id = None
        else:
            _id = json_data[Node._JSON_ID]
            assert type(_id) is int
        # load process_pt
        if Node._JSON_PROCESS_PT not in json_data:
            _logger.error("Missing node process_pt in json description")
            _process_pt = None
        else:
            _process_pt = json_data[Node._JSON_PROCESS_PT]
        return _id, _properties, _process_pt

    @staticmethod
    def from_json(json_data: dict):
        """

        :param json_data: node description
        :return: New node created from the json description
        :rtype: Node
        :raise ValueError: if properties or id or processing_pt missing
        """
        _id, _properties, _process_pt = Node.load_node_info_from_json(json_data)
        if _properties is None or _id is None or _process_pt is None:
            raise ValueError(
                "Unable to create Node from json, core information " "are missing"
            )
        else:
            return Node(id=_id, properties=_properties, processing_pt=_process_pt)

    def __str__(self):
        return "node %s - %s" % (self.id, self._process_pt)


class WorkflowException(Exception):
    def __init__(self, traceBack="", data=None, msg=None):
        if data is None:
            data = {}
        super(WorkflowException, self).__init__(msg)
        self.errorMessage = msg
        self.data = data
        self.traceBack = traceBack


class ErrorHandler(Node):
    """
    TODO
    """

    def __init__(self, processing_pt, id=None, properties=None):
        super(ErrorHandler, self).__init__(
            processing_pt=processing_pt,
            id=id,
            properties=properties,
            error_handler=None,
        )

    def _get_error_handler_json(self):
        return {}



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""utils for the tests"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/04/2020"


class ProcessingClass1:
    """Class to test the workflow. Simply add one to input"""

    def process(self, input):
        input["data"] += 4
        return input

    __call__ = process


class ProcessingClass2:
    """Class to test the workflow. Simply add self.value to input"""

    # define handler has name, type, and handler
    inputs = [["data", dict, "process"]]

    def __init__(self):
        self.value = 1

    def process(self, input):
        input["data"] += self.value
        return input

    def setConfiguration(self, config):
        self.value = config["value"]

    __call__ = process


def test_function(input):
    """Class to test the workflow. Simply add self.value to input"""
    input["data"] += 2
    return input



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

from pypushflow.AbstractActor import AbstractActor


class JoinActor(AbstractActor):
    def __init__(self, parent=None, name="Join actor"):
        AbstractActor.__init__(self, parent=parent, name=name)
        self.numberOfThreads = 0
        self.listInData = []

    def increaseNumberOfThreads(self):
        self.numberOfThreads += 1

    def trigger(self, in_data):
        if in_data is None:
            channel = data = None
        else:
            channel, data = in_data
        self.listInData.append(data)
        if len(self.listInData) == self.numberOfThreads:
            newInData = {}
            for data in self.listInData:
                newInData.update(data)
            for actor in self.listDownStreamActor:
                actor.trigger((channel, data))


class JoinUntilStopSignal(AbstractActor):
    def __init__(self, name):
        self.name = name
        self.listInData = []
        self.listDownStreamActor = []
        self._nprocess_received = 0
        self._nprocess_waited = 0
        self._can_stop = False

    def connect(self, actor):
        self.listDownStreamActor.append(actor)

    def trigger(self, inData):
        if (
            type(inData) is dict
            and "sig_type" in inData
            and inData["sig_type"] == "stop"
        ):
            self._can_stop = True
            self._nprocess_waited = inData["n_process"]
        else:
            self._nprocess_received += 1

        self.listInData.append(inData)
        if self._can_stop and self._nprocess_waited <= self._nprocess_received:
            newInData = {}
            for data in self.listInData:
                newInData.update(data)
            for actor in self.listDownStreamActor:
                actor.trigger(newInData)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "29/05/2017"


from typing import Union
import logging

_logger = logging.getLogger(__name__)


# TODO: next lines should be removed
global next_link_free_id
next_link_free_id = 0


def get_next_link_free_id():
    global next_link_free_id
    _id = next_link_free_id
    next_link_free_id += 1
    return _id


class Link(object):
    """
    Define a link between two node with an execution order. Sink_node will be
    executed after source_node.

    :param  source_node: upstream node
    :type: Union[`.Node`,int]
    :param sink_node: downstream node
    :type: Union[`.Node`,int]
    :param str source_channel: channel name used for connection
    :param str sink_channel: channel name used for connection
    """

    _JSON_SOURCE_CHANNEL = "source_channel"
    _JSON_SINK_CHANNEL = "sink_channel"
    _JSON_SOURCE_NODE_ID = "source_node_id"
    _JSON_SINK_NODE_ID = "sink_node_id"
    _JSON_LINK_ID = "link_id"

    def __init__(
        self,
        source_node,
        sink_node,
        source_channel: str = "default",
        sink_channel: str = "default",
        id=None,
    ):
        self.id = get_next_link_free_id() if id is None else id
        if isinstance(source_node, int):
            self.source_node_id = source_node
        else:
            self.source_node_id = source_node.id

        if isinstance(sink_node, int):
            self.sink_node_id = sink_node
        else:
            self.sink_node_id = sink_node.id

        self.source_channel = source_channel
        self.sink_channel = sink_channel

    def to_json(self) -> dict:
        """

        :return: Link description to the json format
        :rtype: dict
        """
        return {
            self._JSON_LINK_ID: self.id,
            self._JSON_SOURCE_CHANNEL: self.source_channel,
            self._JSON_SINK_CHANNEL: self.sink_channel,
            self._JSON_SOURCE_NODE_ID: self.source_node_id,
            self._JSON_SINK_NODE_ID: self.sink_node_id,
        }

    @staticmethod
    def from_json(json_data: dict):
        """

        :param json_data: link description
        :return: New Link created from the json description
        :rtype: Link
        :raise ValueError: if sink or source channel missing or if link id
                           missing or if sink or source node missing
        """
        # load link id
        if Link._JSON_LINK_ID not in json_data:
            _id = None
            _logger.error("Missing link id information")
        else:
            _id = json_data[Link._JSON_LINK_ID]
        # load sink channel
        if Link._JSON_SINK_CHANNEL not in json_data:
            sink_channel = None
            _logger.error("Missing sink channel information")
        else:
            sink_channel = json_data[Link._JSON_SINK_CHANNEL]
        # load source channel
        if Link._JSON_SOURCE_CHANNEL not in json_data:
            source_channel = None
            _logger.error("Missing source channel information")
        else:
            source_channel = json_data[Link._JSON_SOURCE_CHANNEL]
        # load sink node id
        if Link._JSON_SINK_NODE_ID not in json_data:
            sink_node_id = None
            _logger.error("Missing source node id information")
        else:
            sink_node_id = json_data[Link._JSON_SINK_NODE_ID]
        # load source node id
        if Link._JSON_SOURCE_NODE_ID not in json_data:
            source_node_id = None
            _logger.error("Missing source node id information")
        else:
            source_node_id = json_data[Link._JSON_SOURCE_NODE_ID]

        if (
            sink_channel is None
            or source_channel is None
            or _id is None
            or source_node_id is None
            or sink_node_id is None
        ):
            raise ValueError("Missing core information for creating a Link")
        else:
            return Link(
                id=_id,
                sink_channel=sink_channel,
                source_channel=source_channel,
                source_node=source_node_id,
                sink_node=sink_node_id,
            )

    def __str__(self):
        return "node %s: source:(%s, %s), sink:(%s, %s)" % (
            self.id,
            self.source_node_id,
            self.source_channel,
            self.sink_node_id,
            self.sink_channel,
        )



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

from pypushflow.AbstractActor import AbstractActor

import logging

logger = logging.getLogger("pypushflow")


class RouterActor(AbstractActor):
    def __init__(
        self, parent, errorHandler=None, name="Router", itemName=None, listPort=[]
    ):
        AbstractActor.__init__(self, parent, name)
        self.errorHandler = errorHandler
        self.name = name
        self.itemName = itemName
        self.listPort = listPort
        self.dictValues = {}

    def connect(self, actor, expectedValue="other"):
        if expectedValue != "other" and not expectedValue in self.listPort:
            raise RuntimeError(
                "Port {0} not defined for router actor {1}!".format(
                    expectedValue, self.name
                )
            )
        if expectedValue in self.dictValues:
            self.dictValues[expectedValue].append(actor)
        else:
            self.dictValues[expectedValue] = [actor]

    def trigger(self, inData):
        logger.debug('In router actor "{0}"'.format(self.name))
        listActor = None
        if self.itemName in inData and not inData[self.itemName] in [
            None,
            "None",
            "null",
        ]:
            logger.debug(
                'In router actor "{0}", itemName {1} in inData'.format(
                    self.name, self.itemName
                )
            )
            value = inData[self.itemName]
            logger.debug('In router actor "{0}", value = {1}'.format(self.name, value))
            if not isinstance(value, dict) and value in self.dictValues:
                listActor = self.dictValues[value]
        if listActor is None:
            logger.debug('In router actor "{0}", actor is None')
            if "other" in self.dictValues:
                listActor = self.dictValues["other"]
            else:
                raise RuntimeError(
                    'No "other" port for router actor "{0}"'.format(self.name)
                )
        for actor in listActor:
            logger.debug(
                'In router actor "{0}", triggering actor "{1}"'.format(
                    self.name, actor.name
                )
            )
            actor.trigger(inData)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2019-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/04/2020"


import logging
import pkgutil
import inspect
import importlib

_logger = logging.getLogger(__name__)


def get_registered_add_ons():
    """Return the list of registered add-on"""
    try:
        import ppfaddon
    except ImportError:
        _logger.info("no add on found")
        return []
    else:
        modules = []
        for importer, modname, ispkg in pkgutil.iter_modules(ppfaddon.__path__):
            if ispkg:
                modules.append(modname)
        return modules


def _check_module_for_class(module, depth=0):
    classes = []
    for m in inspect.getmembers(module):

        if inspect.isclass(m[1]):
            if m[1].__module__ != "pypushflow.addon.classes":
                classes.append(m[1])

        if inspect.ismodule(m[1]):
            if inspect.isclass(m[1]):
                classes.append(m[1])
            elif hasattr(m[1], "__path__"):
                for importer, modname, ispkg in pkgutil.iter_modules(m[1].__path__):
                    module_name = module.__name__
                    submod_name = ".".join((module_name, m[0], modname))
                    try:
                        mod = importlib.import_module(submod_name)
                    except Exception:
                        pass
                    else:
                        classes.extend(
                            _check_module_for_class(module=mod, depth=depth + 1)
                        )
    return classes


def get_registered_add_ons_classes():
    """Return a dictionary with the list of registered add-on classes for each
    add-on"""
    classes = {}

    add_on_modules = get_registered_add_ons()
    for module_name in add_on_modules:
        package_name = ".".join(("ppfaddon", module_name))
        try:
            module = __import__(package_name)
        except:
            _logger.error("fail to import {}".format(package_name))
            return []

        classes[module_name] = _check_module_for_class(module=module)
    return classes


def is_add_on_class_relative_to(add_on_class, myclass):
    """Return the list of registered add-on classes"""
    return isinstance(myclass, add_on_class.target)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import logging
from pypushflow.addon import utils
from pypushflow.addon.classes import BaseActorAddOn


logger = logging.getLogger("pypushflow")


class AbstractActor(object):
    """
    TODO
    """

    def __init__(self, parent=None, name=None):
        if name is None:
            raise RuntimeError("Actor name is None!")
        self.name = name
        self.listDownStreamActor = []
        self.parent = parent
        self.actorId = None

        self._add_ons = []
        for add_on_class in self._getAddOnsClasses():
            self._add_ons.append(add_on_class())

    def connect(self, actor):
        logger.debug(
            'Connecting actor "{0}" to actor "{1}"'.format(self.name, actor.name)
        )
        self.listDownStreamActor.append(actor)

    def trigger(self, inData):
        for actor in self.listDownStreamActor:
            logger.debug(
                'In actor "{0}", triggering actor "{1}"'.format(self.name, actor.name)
            )
            self._process_pre_trigger_add_on(inData)
            actor.trigger(inData)
            self._process_post_trigger_add_on()

    def getActorPath(self):
        return self.parent.getActorPath()

    def get_addons(self):
        """Return the list of add-on that will be apply to this actor"""
        raise NotImplementedError()

    def _process_post_trigger_add_on(self):
        for add_on in self._add_ons:
            add_on.post_trigger_action(actor=self)

    def _process_pre_trigger_add_on(self, in_data):
        for add_on in self._add_ons:
            add_on.pre_trigger_action(actor=self, in_data=in_data)

    def _getAddOnsClasses(self):
        add_ons = []
        for _, classes in utils.get_registered_add_ons_classes().items():
            for class_ in classes:
                import inspect

                if BaseActorAddOn in (inspect.getmro(class_)):
                    add_ons.append(class_)
        return add_ons



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import os
import logging
import unittest
from pypushflow.test.TestWorkflow import Workflow1
from pypushflow.test.TestWorkflow import Workflow2
from pypushflow.test.TestWorkflow import Workflow3

try:
    import pymongo
except ImportError:
    _has_pymongo = False
else:
    _has_pymongo = True


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("testWorkflow")


@unittest.skipUnless(_has_pymongo, "need pymongo install")
class TestWorkflows(unittest.TestCase):
    def setUp(self):
        os.environ[
            "PYPUSHFLOW_MONGOURL"
        ] = "mongodb://pybes:pybes@linsvensson.esrf.fr:27017/pybes"
        os.environ["PYPUSHFLOW_INITIATOR"] = "TestWorkflows"

    def test_Workflow1(self):
        testWorkflow1 = Workflow1("Test workflow 1")
        inData = {"name": "Jerry"}
        out_data = testWorkflow1.run(inData)
        self.assertIsNotNone(out_data)
        self.assertEqual(out_data["reply"], "Hello Jerry!")

    def test_Workflow2(self):
        testWorkflow2 = Workflow2("Test workflow 2")
        inData = {"name": "Tom"}
        out_data = testWorkflow2.run(inData)
        self.assertIsNotNone(out_data)
        self.assertTrue("WorkflowException" in out_data)

    def test_Workflow3(self):
        testWorkflow3 = Workflow3("Test workflow 3")
        inData = {"name": "Cat"}
        out_data = testWorkflow3.run(inData)
        self.assertIsNotNone(out_data)
        self.assertEqual(out_data["reply"], "Hello Cat!")

    def test_Workflow4(self):
        testWorkflow4 = Workflow2("Test workflow 4")
        inData = {"name": "Dog"}
        out_data = testWorkflow4.run(inData)
        self.assertIsNotNone(out_data)
        self.assertTrue("WorkflowException" in out_data)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestWorkflows,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson", "H. Payno"]
__license__ = "MIT"
__date__ = "28/05/2019"


import pprint
import logging
from pypushflow.representation import Scheme
from pypushflow.PythonActor import PythonActor as ActorFactory
from pypushflow.StartActor import StartActor
from pypushflow.StopActor import StopActor
from pypushflow.JoinActor import JoinUntilStopSignal
from pypushflow.addon import utils
from pypushflow.addon.classes import BaseWorkflowAddOn


logger = logging.getLogger("pypushflow")


class _BaseWorkflow:
    def __init__(self, configuration):
        self._configuration = configuration or {}
        self._add_ons = []
        for add_on_class in self._getAddOnsClasses():
            self._add_ons.append(
                add_on_class(workflow=self, configuration=self._configuration)
            )

    def _getAddOnsClasses(self):
        add_ons = []
        for _, classes in utils.get_registered_add_ons_classes().items():
            for class_ in classes:
                import inspect

                if BaseWorkflowAddOn in (inspect.getmro(class_)):
                    add_ons.append(class_)
        return add_ons


class Workflow(_BaseWorkflow):
    """TODO"""

    def __init__(self, name):
        super(Workflow, self).__init__()
        self.name = name
        self.listOnErrorActor = []

    def connectOnError(self, actor):
        logger.debug(
            "In connectOnError in subModule {0}, actor name {1}".format(
                self.name, actor.name
            )
        )
        self.listOnErrorActor.append(actor)

    def triggerOnError(self, inData):
        logger.debug(pprint.pformat(inData))
        for onErrorActor in self.listOnErrorActor:
            logger.debug(onErrorActor.trigger)
            onErrorActor.trigger(inData)

    def getActorPath(self):
        return "/" + self.name


class ProcessableWorkflow(_BaseWorkflow):
    """Define a workflow that can be executed

    :param scheme: the workflow scheme
    :param configuration: some configuration / settings that can be pass to
                          the add-on.
    """

    def __init__(self, scheme, configuration=None):
        super(ProcessableWorkflow, self).__init__(configuration=configuration)
        assert isinstance(scheme, Scheme)
        self._representation = scheme
        # first load node handlers if any
        scheme.load_handlers()

        self._actor_factory = {}
        for node in self._representation.nodes:
            name = "-".join((str(node.id), node._process_pt))
            self._actor_factory[node] = ActorFactory(
                parent=None, name=name, node=node, errorHandler=None
            )

        # deal with connect
        for node in self._representation.nodes:
            actor_factory = self._actor_factory[node]
            for downstream_node in node.downstream_nodes:
                downstream_actor_factory = self._actor_factory[downstream_node]
                actor_factory.connect(downstream_actor_factory)

        # add start actor
        self._start_actor = StartActor()
        for node in self._representation.start_nodes():
            actor_factory = self._actor_factory[node]
            self._start_actor.connect(actor_factory)

        def connect_finals_nodes(actor):
            # add end actor
            for node in self._representation.final_nodes():
                actor_factory = self._actor_factory[node]
                actor_factory.connect(actor)

        self._end_actor = StopActor()

        if self.has_final_join():
            self._join_actor = JoinUntilStopSignal("stop join")
            connect_finals_nodes(self._join_actor)
            self._join_actor.connect(self._end_actor)
        else:
            connect_finals_nodes(self._end_actor)

    def has_final_join(self):
        """True if we need to send a 'end' signal before closing the workflow
        This is needed for DataList and DataWatcher
        """
        for node in self._representation.nodes:
            if node.need_stop_join:
                return True
        return False


class Converter:
    """
    Write a Workflow to a python file which can be executed later
    """

    def __init__(self, workflow, output_file):
        if not isinstance(workflow, _BaseWorkflow):
            raise TypeError("workflow should be an instance of `_BaseWorkflow`")
        self.workflow = workflow
        self.output_file = output_file

    def process(self):
        self._write_import()
        self._write_util_functions()
        self._write_main_section()
        self._write_processes_creation()
        self._write_processing()
        self._close_main_section()

    def _write_main_section(self):
        with open(self.output_file, "a") as file_:
            file_.write("\n\n")
            file_.write("def main(input_data, channel):\n")

    def _close_main_section(self):
        with open(self.output_file, "a") as file_:
            file_.write("\n\n")

    def _write_import(self):
        with open(self.output_file, "a") as file_:
            file_.write("\n")
            for node in self.workflow._representation.nodes:
                class_name, mod_name = node.get_class_name_and_module_name()
                file_.write("import {}\n".format(mod_name))
            # create logger
            file_.write(
                "{}\n{}\n".format(
                    "import logging", "_logger = logging.getLogger(__name__)"
                )
            )

    def _write_processes_creation(self):
        with open(self.output_file, "a") as file_:
            file_.write("\n")
            for node in self.workflow._representation.nodes:
                class_name, mod_name = node.get_class_name_and_module_name()
                file_.write(
                    "    process_{} = {}.{}()\n".format(node.id, mod_name, class_name)
                )
                if hasattr(node.class_instance, "set_properties"):
                    # filter some orange properties
                    properties = node.properties
                    for param in (
                        "controlAreaVisible",
                        "savedWidgetGeometry",
                        "__version__",
                        "libraryListSource",
                    ):
                        if param in properties:
                            del properties[param]
                    file_.write(
                        "    process_{}.set_properties({})\n".format(
                            node.id, properties
                        )
                    )

    def _write_processing(self):
        for _, link in self.workflow._representation.links.items():
            print(link)
            self._write_connection(
                sink_channel=link.sink_channel,
                source_channel=link.source_channel,
                sink_node_id=link.sink_node_id,
                source_node_id=link.source_node_id,
            )

        with open(self.output_file, "a") as file_:
            file_.write("\n# start processing\n")
            for node in self.workflow._representation.start_nodes():
                self._write_starter(node)

    def _write_starter(self, node):
        with open(self.output_file, "a") as file_:
            file_.write("    trigger(process_{}, input_data, channel)".format(node.id))

    def _write_connection(
        self, sink_channel, source_channel, sink_node_id, source_node_id
    ):
        with open(self.output_file, "a") as file_:
            file_.write(
                '    connect(process_{}, "{}", process_{}, '
                '"{}")\n'.format(
                    source_node_id, source_channel, sink_node_id, sink_channel
                )
            )

    def _write_util_functions(self):
        with open(self.output_file, "a") as file_:
            file_.write(self._get_utils_functions())

    def _get_utils_functions(self):
        return """
connections = {}


def get_output_channel_name(class_inst, output):
    assert hasattr(class_inst, 'outputs')
    for output_ in class_inst.outputs:
            output_name, output_type_, output_handler = output_[:3]
            if isinstance(output, output_type_):
                return output_name
    return None

def get_handler(class_inst, channel_name):
    '''Find the associate handler for channel `channel_name`'''
    if hasattr(class_inst, 'inputs'):
        for input_ in class_inst.inputs:
            input_name, input_type, input_handler = input_[:3]
            if input_name == channel_name:
                return input_handler
        return None
    return None


def connect(source_process, source_channel, sink_process, sink_channel):
    if (source_process, source_channel) not in connections:
        connections[(source_process, source_channel)] = []
    handler = get_handler(sink_process, channel_name=sink_channel)
    if handler is None:
        return TypeError('{} channel is not managed by {}'.format(sink_channel,
                                                                  sink_process))
    connections[source_process, source_channel].append((sink_process, handler))
    '''Register for each cuple process / outputtype the process and handler
    to launch when processes end'''


def trigger(process, input_data, channel_name):
    _logger.info('{} has been trigger on channel {} with input {}'
                 ''.format(process, channel_name, input_data))
    if isinstance(process, IgnoreProcess):
        output_data = input_data
        output_channel_name = channel_name
    else:
        handler = get_handler(class_inst=process, channel_name=channel_name)
        if handler is None:
            raise ValueError('Fail to find handler for {} on channel {}'
                             ''.format(process, channel_name))
        assert hasattr(process, handler)
        output_data = getattr(process, handler)(input_data)
        output_channel_name = get_output_channel_name(process, output_data)

    if (process, output_channel_name) in connections:
        for downstream in connections[(process, output_channel_name)]:
            sink_process, handler = downstream
            trigger(sink_process, output_data, output_channel_name)
        """
